#include <Servo.h>
#include "IRremote.h"

// digital outputs
#define RX 0              // RX serial communication
#define TX 1              // TX serial communication
#define SPEAKJET_TX 1     // speakjet TX serial communication
#define SPEAKJET_RESET 2 // speakjet reset signal
#define SERVO 3           // servo motor for the head
#define E1 5              // M1 left motor speed control
#define E2 6              // M2 right motor speed control
#define M1 7              // M1 left motor direction control
#define M2 8              // M2 right motor direction control
#define LED_STORE 9       // store-signal for shift register
#define LED_SHIFT 10      // clock signal for shift register
#define IR_INPUT 11       // input for IR remote control
#define LED_DS 12          // data for shift register
#define STATUS_LED 13     // led on port 13

// analog inputs
#define DISTANCE_SENSOR 0 // IR distance sensor
#define LDR_L 4 // LDR left
#define LDR_R 5 // LDR right

// speakjet constants
#define P_VOLUME 20
#define P_SPEED 21
#define P_BEND 23
#define P_PITCH 22
#define VOLUME 127
#define SPEED 120
#define BEND 9

// other constants
#define MOTOR_SPEED 100         // speed of the robot 0-255
#define COLLISION_DISTANCE 500  // if distance is higher than this value, turn right
#define COLLISION_DISTANCE_2 600  // if distance is higher than this value, go backward
#define NIGHT_LIMIT 40          // minimum value to detect night from LDR sensor
#define DAY_LIMIT 60            // minimum value to detect day from LDR sensor
#define MAX_WALK_TIME 30000     // time to walk around

// robot states
#define STATE_SLEEP 0
#define STATE_CHANGE_FACE 1
#define STATE_PLAY_SOUND 2
#define STATE_MOVE_FORWARD 3
#define STATE_MOVE_BACKWARD 4
#define STATE_TURN_LEFT 5
#define STATE_TURN_RIGHT 6
#define STATE_LOOK_AROUND 7
#define STATE_STOP 8
#define STATE_LOOK_TO_CLOSEST 9
#define STATE_NIGHT_SLEEP 10

// robot state information
int robotState;
boolean remoteControl;
unsigned long walkStopTime;
boolean walking;

// other variables
unsigned shiftValue;
Servo headServo;
IRrecv irrecv(IR_INPUT);

// sounds
#define P_IY 128
#define P_IH 129
#define P_EY 130
#define P_EH 131
#define P_AY 132
#define P_AX 133
#define P_UX 134
#define P_OH 135
#define P_AW 136
#define P_OW 137
#define P_UH 138
#define P_UW 139
#define P_MM 140
#define P_NE 141
#define P_NO 142
#define P_NGE 143
#define P_NGO 144
#define P_LE 145
#define P_LO 146
#define P_WW 147
#define P_RR 148
#define P_IYRR 149
#define P_EYRR 150
#define P_AXRR 151
#define P_AWRR 152
#define P_OWRR 153
#define P_EYIY 154
#define P_OHIY 155
#define P_OWIY 156
#define P_OHIH 157
#define P_IYEH 158
#define P_EHLL 159
#define P_IYUW 160
#define P_AXUW 161
#define P_IHWW 162
#define P_AYWW 163
#define P_OWWW 164
#define P_JH 165
#define P_VV 166
#define P_ZZ 167
#define P_ZH 168
#define P_DH 169
#define P_BE 170
#define P_BO 171
#define P_EB 172
#define P_OB 173
#define P_DE 174
#define P_DO 175
#define P_ED 176
#define P_OD 177
#define P_GE 178
#define P_GO 179
#define P_EG 180
#define P_OG 181
#define P_CH 182
#define P_HE 183
#define P_HO 184
#define P_WH 185
#define P_FF 186
#define P_SE 187
#define P_SO 188
#define P_SH 189
#define P_TH 190
#define P_TT 191
#define P_TU 192
#define P_TS 193
#define P_KE 194
#define P_KO 195
#define P_EK 196
#define P_OK 197
#define P_PE 198
#define P_PO 199
char aan_tafel[] = {P_PITCH, 156, P_AY, P_AY, P_NE, 6, P_PITCH, 123, P_TU, P_AY, P_AY, P_AY, P_FF, P_UX, P_LE, 0};
char noam[] = {6, P_NO, P_OW, P_OWWW, P_AW, P_AW, P_MM, P_MM, 0};
char sarah[] =  {6, P_SE, P_AW, P_AW, P_AWRR, P_AW, P_AW, 0};
char amber[] = {6, P_AW, P_MM, P_BE, P_UX, P_RR, 0};
char kris[] = {6, P_KO, P_RR, P_IY, P_SE, 0};
char komen_eten[] = {6, P_PITCH, 93, P_KO, P_OWWW, P_OWWW, P_MM, P_UX, P_NE, 6, P_PITCH, 156, P_EY, P_EY, P_EY, P_PITCH, 123, P_TT, P_UX, P_NE, 0};
char robot_sounds[] = {200, 201, 202, 203, 220, 221, 222, 0};

// motor control
void stopMotors() {
  digitalWrite(E1, LOW);   
  digitalWrite(E2, LOW);
  digitalWrite(M1, LOW);
  digitalWrite(M2, LOW);  
}   
void advance() {
  //analogWrite(E1,MOTOR_SPEED);
  digitalWrite(E1, HIGH);
  digitalWrite(M1,HIGH);    
  //analogWrite(E2,MOTOR_SPEED);    
  digitalWrite(E2, HIGH);
  digitalWrite(M2,HIGH);
}  
void backOff () {
  //analogWrite(E1,MOTOR_SPEED);
  digitalWrite(E1, HIGH);
  digitalWrite(M1,LOW);   
  //analogWrite(E2,MOTOR_SPEED);
  digitalWrite(E2, HIGH);  
  digitalWrite(M2,LOW);
}
void turnLeft () {
  //analogWrite(E1,MOTOR_SPEED);
  digitalWrite(E1, HIGH);
  digitalWrite(M1,LOW);    
  //analogWrite(E2,MOTOR_SPEED);
  digitalWrite(E2, HIGH);  
  digitalWrite(M2,HIGH);
}
void turnRight () {
  //analogWrite(E1,MOTOR_SPEED);
  digitalWrite(E1, HIGH);
  digitalWrite(M1,HIGH);    
  //analogWrite(E2,MOTOR_SPEED);
  digitalWrite(E2, HIGH);  
  digitalWrite(M2,LOW);
}

// face light control
void setFaceLight(int bitNumber, int value) {
  int mask = 1 << bitNumber;
  if (value == HIGH) {
    shiftValue |= mask;
  } 
  else {
    mask = ~mask;
    shiftValue &= mask;
  } 
}
void showFaceLights() {
  unsigned t = shiftValue;
  for(int i=0; i<16; i++) {
    if ((t & 1)==1) digitalWrite(LED_DS, HIGH); 
    else digitalWrite(LED_DS, LOW);
    t = t >> 1;
    digitalWrite(LED_SHIFT, HIGH);
    delay(1);
    digitalWrite(LED_SHIFT, LOW);
  }
  digitalWrite(LED_STORE, HIGH);
  delay(1);
  digitalWrite(LED_STORE, LOW);
  delay(1);
}

// sleep state
unsigned long sleepStopTime;
void sleepInit() {
  robotState = STATE_SLEEP;
  stopMotors();
  sleepStopTime = random(3000, 7000) + millis();
}
void sleepRun() {
  unsigned long time = millis();
  if (time > sleepStopTime) {
    sleepStop();
    int r = random(0,4);
    if (r == 0) {
      changeFaceInit();
    } else if (r == 1) {
      lookAroundInit();
    } else if (r == 2) {
      playSoundInit();
    } else {
      walkStopTime = millis() + MAX_WALK_TIME;
      walking = true;
      moveForwardInit();
    }
  }
}
void sleepStop() {
  setFaceLight(13, HIGH);
  setFaceLight(14, HIGH);
  setFaceLight(15, HIGH);
  showFaceLights();
}

// change face state
int numberOfLights;
int currentNumberOfLights;
unsigned long faceChangeStartTime;
void changeFaceInit() {
  shiftValue = 0;
  setFaceLight(6, HIGH);
  setFaceLight(15, HIGH);
  setFaceLight(14, HIGH);
  setFaceLight(13, HIGH);
  showFaceLights();
  robotState = STATE_CHANGE_FACE;
  numberOfLights = random(0, 7);
  currentNumberOfLights = 0;
  faceChangeStartTime = millis();
}
void changeFaceRun() {
  unsigned long numberOfLightsToShow = (millis() - faceChangeStartTime) / 200;
  if (numberOfLightsToShow > (numberOfLights+1)) {
    changeFaceStop();
    if (remoteControl) {
      stopInit();
    } else {
      playSoundInit();
    }
  } else {
    if (currentNumberOfLights != numberOfLightsToShow) {
      for(int i=0; i<numberOfLightsToShow; i++) {
        setFaceLight(6+i, HIGH);
        setFaceLight(6-i, HIGH);
      }
      showFaceLights();
      currentNumberOfLights = numberOfLightsToShow;
    }
  }
}
void changeFaceStop() {
}

// move forward state
unsigned long moveTimeOut;
void moveForwardInit() {
  robotState = STATE_MOVE_FORWARD;
  moveTimeOut = millis() + 5000;
  headServo.attach(SERVO, 600, 2200);
  headServo.write(90);
  delay(500);
  headServo.detach();
  advance();
}
void moveForwardRun() {
  int distance = analogRead(DISTANCE_SENSOR);
  if (millis() > moveTimeOut) {
    distance = COLLISION_DISTANCE + 1;
    moveTimeOut = millis() + 5000;
    if (!walking) {
      moveForwardStop();
      if (remoteControl) {
        stopInit();
      } else {
        sleepInit();
      }
      return;
    }
  }
  if (distance > COLLISION_DISTANCE) {
    moveForwardStop();
    if (distance > COLLISION_DISTANCE_2) {
      moveBackwardInit();
    } else {
      turnRightInit();
    }
  } else if (millis() > walkStopTime) {
    moveForwardStop();
    if (remoteControl) {
      stopInit();
    } else {
      sleepInit();
    }
  }
}
void moveForwardStop() {
  stopMotors();
}

// move backward state
unsigned long moveBackwardTimeOut;
void moveBackwardInit() {
  robotState = STATE_MOVE_BACKWARD;
  moveBackwardTimeOut = 500 + millis();
  backOff();
}
void moveBackwardRun() {
  if (millis() > moveBackwardTimeOut) {
    moveBackwardStop();
    if (remoteControl && !walking) {
      stopInit();
    } else {
      turnLeftInit();
    }
  }
}
void moveBackwardStop() {
  stopMotors();
}

// turn left state
unsigned long turnStopTime;
void turnLeftInit() {
  turnStopTime = 500 + millis();
  robotState = STATE_TURN_LEFT;
  turnLeft();
}
void turnLeftRun() {
  if (millis() > turnStopTime) {
    turnLeftStop();
    if (remoteControl && !walking) {
      stopInit();
    } else {
      moveForwardInit();
    }
  }
}
void turnLeftStop() {
  stopMotors();
}

// turn right state
void turnRightInit() {
  turnStopTime = 500 + millis();
  robotState = STATE_TURN_RIGHT;
  turnRight();
}
void turnRightRun() {
  if (millis() > turnStopTime) {
    turnRightStop();
    if (remoteControl && !walking) {
      stopInit();
    } else {
      moveForwardInit();
    }
  }
}
void turnRightStop() {
  stopMotors();
}

// look around state
int headPosition;
int closestAngle;
int closestDistance;
int farestAngle;
int farestDistance;
void lookAroundInit() {
  robotState = STATE_LOOK_AROUND;
  headServo.attach(SERVO, 600, 2200);
  headServo.write(30);
  headPosition = 30;
  closestDistance = 0;
  closestAngle = 90;
  farestDistance = 2000;
  farestAngle = 90;
  delay(1000);
}
void lookAroundRun() {
  headPosition+=5;
  if (headPosition > 150) {
    lookAroundStop();
    lookToClosestInit();
    return;
  }
  headServo.write(headPosition);
  delay(50);
  int distance = analogRead(DISTANCE_SENSOR);
  if (distance < farestDistance) {
    farestDistance = distance;
    farestAngle = headPosition;
  }
  if (distance > closestDistance) {
    closestDistance = distance;
    closestAngle = headPosition;
  }
}
void lookAroundStop() {
}

// stop state
void stopInit() {
  robotState = STATE_STOP;
  stopMotors();
  setFaceLight(13, LOW);
  showFaceLights();
}
void stopRun() {
}
void stopStop() {
  setFaceLight(13, HIGH);
  showFaceLights();
}

// play sound state
unsigned long playSoundEndTime;
void playSoundInit() {
  robotState = STATE_PLAY_SOUND;
  int r = random(0,100);
  playSoundEndTime = millis() + 1000;
  if (r < 50) {
    int numberOfSounds = 6;
    playSoundEndTime = millis() + numberOfSounds*500;
    char sounds[7];
    for(int i=0; i<numberOfSounds; i++) {
      sounds[i] = random(200, 240);
    }
    sounds[numberOfSounds] = 0;
    Serial.print(sounds);
  } else if (r < 70) {
    Serial.print(noam);
  } else if (r < 90) {
    Serial.print(sarah);
  } else {
    Serial.print(komen_eten);
  }
}
void playSoundRun() {
  if (millis() > playSoundEndTime) {
    playSoundStop();
    if (remoteControl) {
      stopInit();
    } else {
      int r = random(0,4);
      if (r == 0) {
        changeFaceInit();
      } else if (r == 1) {
        lookAroundInit();
      } else if (r == 2) {
        sleepInit();
      } else {
        walkStopTime = millis() + MAX_WALK_TIME;
        walking = true;
        moveForwardInit();
      }
    }
  }
}
void playSoundStop() {
}

// look to closest state
unsigned long lookToClosestStopTime;
void lookToClosestInit() {
  robotState = STATE_LOOK_TO_CLOSEST;
  headServo.write(closestAngle-20);
  delay(1000);
  headServo.detach();
  lookToClosestStopTime = millis() + 1;
}
void lookToClosestRun() {
  if (millis() > lookToClosestStopTime) {
    lookToClosestStop();
    if (remoteControl) {
      stopInit();
    } else {
      changeFaceInit();
    }
  }
}
void lookToClosestStop() {
}

void nightSleepInit() {
  robotState = STATE_NIGHT_SLEEP;
  stopMotors();
  shiftValue=0;
  showFaceLights();
  headServo.attach(SERVO, 600, 2200);
  headServo.write(90);
  delay(1000);
  headServo.detach();
}
void nightSleepRun() {
  int luminationLeft = analogRead(LDR_L);
  int luminationRight = analogRead(LDR_R);
  if (luminationLeft>DAY_LIMIT || luminationRight>DAY_LIMIT) {
    nightSleepStop();
    changeFaceInit();
  }
}
void nightSleepStop() {
}

void setup(void) 
{
  pinMode(SPEAKJET_TX, OUTPUT);
  pinMode(SERVO, OUTPUT);
  pinMode(E1, OUTPUT);
  pinMode(M1, OUTPUT);
  pinMode(E2, OUTPUT);
  pinMode(M2, OUTPUT);
  pinMode(LED_DS, OUTPUT);
  pinMode(LED_SHIFT, OUTPUT);
  pinMode(LED_STORE, OUTPUT);
  pinMode(SPEAKJET_RESET, OUTPUT);
  //pinMode(STATUS_LED, OUTPUT);

  stopMotors();
  
  shiftValue = 0;

  //headServo.attach(SERVO, 600, 2200);
  digitalWrite(LED_STORE, LOW);
  digitalWrite(LED_SHIFT, LOW);
  
  irrecv.enableIRIn();
  remoteControl = false;
  
  walking = true;
  
  Serial.begin(9600);
  digitalWrite(SPEAKJET_RESET, LOW);
  delay(100);
  digitalWrite(SPEAKJET_RESET, HIGH);
  delay(100);
  char initialise[] = {P_VOLUME, VOLUME, P_SPEED, SPEED, P_BEND, BEND, 0};
  Serial.print(initialise);

  sleepInit();
}

// stops current state
void stopCurrentState() {
  if (robotState == STATE_SLEEP) {
    sleepStop();
  } else if (robotState == STATE_CHANGE_FACE) {
    changeFaceStop();
  } else if (robotState == STATE_MOVE_FORWARD) {
    moveForwardStop();
  } else if (robotState == STATE_MOVE_BACKWARD) {
    moveBackwardStop();
  } else if (robotState == STATE_TURN_LEFT) {
    turnLeftStop();
  } else if (robotState == STATE_TURN_RIGHT) {
    turnRightStop();
  } else if (robotState == STATE_LOOK_AROUND) {
    lookAroundStop();
  } else if (robotState == STATE_STOP) {
    stopStop();
  } else if (robotState == STATE_PLAY_SOUND) {
    playSoundStop();
  } else if (robotState == STATE_LOOK_TO_CLOSEST) {
    lookToClosestStop();
  } else if (robotState == STATE_NIGHT_SLEEP) {
    nightSleepStop();
  }
}

// handles a command from the IR remote control
// command contains an integer code from the remote control
void handleIRCommand(unsigned long command) {
  // power: go to sleep
  if (command == 0xFD00FF) {
    remoteControl = false;
    stopCurrentState();
    sleepInit();
  }
  // 1: change face
  if (command == 0xFD08F7) {
    remoteControl = true;
    stopCurrentState();
    changeFaceInit();
  }
  // 2: play sound
  if (command == 0xFD8877) {
    remoteControl = true;
    stopCurrentState();
    playSoundInit();
  }
  // 3: look around
  if (command == 0xFD48B7) {
    remoteControl = true;
    stopCurrentState();
    lookAroundInit();
  }
  // 4: wandel rond
  if (command == 0xFD28D7) {
    remoteControl = true;
    stopCurrentState();
    walkStopTime = millis() + MAX_WALK_TIME;
    walking = true;
    moveForwardInit();
  }
  // boven
  if (command == 0xFD50AF) {
    remoteControl = true;
    stopCurrentState();
    walkStopTime = millis() + MAX_WALK_TIME;
    walking = false;
    moveForwardInit();
  }
  // onder
  if (command == 0xFD10EF) {
    remoteControl = true;
    stopCurrentState();
    walking = false;
    moveBackwardInit();
  }
  // links
  if (command == 0xFD20DF) {
    remoteControl = true;
    stopCurrentState();
    walking = false;
    turnLeftInit();
  }
  // rechts
  if (command == 0xFD609F) {
    remoteControl = true;
    stopCurrentState();
    walking = false;
    turnRightInit();
  }
  // play/pauze
  if (command == 0xFDA05F) {
    remoteControl = true;
    stopCurrentState();
    stopInit();
  }
}

// checks if an IR command was received and handles it if necessary
void checkIRCommand() {
   decode_results results;
   if (irrecv.decode(&results)) {
     handleIRCommand(results.value);
     irrecv.resume();
   }
}

// check if surrounding is dark
void checkLumination() {
  int luminationLeft = analogRead(LDR_L);
  int luminationRight = analogRead(LDR_R);
  if (luminationLeft<NIGHT_LIMIT && luminationRight<NIGHT_LIMIT) {
    stopCurrentState();
    nightSleepInit();
  }
}

void loop() {
  //delay(50);
  checkIRCommand();
  //checkLumination();
  if (robotState == STATE_SLEEP) {
    sleepRun();
  } else if (robotState == STATE_CHANGE_FACE) {
    changeFaceRun();
  } else if (robotState == STATE_MOVE_FORWARD) {
    moveForwardRun();
  } else if (robotState == STATE_MOVE_BACKWARD) {
    moveBackwardRun();
  } else if (robotState == STATE_TURN_LEFT) {
    turnLeftRun();
  } else if (robotState == STATE_TURN_RIGHT) {
    turnRightRun();
  } else if (robotState == STATE_LOOK_AROUND) {
    lookAroundRun();
  } else if (robotState == STATE_STOP) {
    stopRun();
  } else if (robotState == STATE_PLAY_SOUND) {
    playSoundRun();
  } else if (robotState == STATE_LOOK_TO_CLOSEST) {
    lookToClosestRun();
  } else if (robotState == STATE_NIGHT_SLEEP) {
    nightSleepRun();
  }
}

